@@ -37,20 +37,24 @@ may be strictly algorithmic, may interact with human subjects, or may be embedde
 sites.
 """
 
-__version__ = "1.1.2"
+__version__ = "1.1.4"
 
 if "dev" in __version__:
     print("PyACTUp version", __version__)
 
 import collections.abc as abc
+import csv
+import io
 import math
-import numbers
+import numpy as np
 import operator
 import pylru
 import random
 import sys
 
 from contextlib import contextmanager
+from numbers import Number
+from prettytable import PrettyTable
 from warnings import warn
 
 __all__ = ("Memory", "set_similarity_function", "use_actr_similarity")
@@ -63,8 +67,10 @@ DEFAULT_RETRIEVAL_TIME_INCREMENT = 0
 
 MINIMUM_TEMPERATURE = 0.01
 
-TRANSCENDENTAL_CACHE_SIZE = 1000
+LN_CACHE_SIZE = 1000
 SIMILARITY_CACHE_SIZE = 10_000
+NOISE_VALUES_SIZE = 1000
+MAXIMUM_RANDOM_SEED = 2**62
 
 class Memory(dict):
     """A cognitive entity containing a collection of learned things, its chunks.
@@ -117,6 +123,10 @@ class Memory(dict):
         self._learning_time_increment = learning_time_increment
         self._retrieval_time_increment = retrieval_time_increment
         self._activation_history = None
+        # Initialize the noise RNG from the parent Python RNG, in case the latter gets seeded for determinancy.
+        self._rng = np.random.default_rng([random.randint(0, MAXIMUM_RANDOM_SEED) for i in range(16)])
+        self._noise_values = None
+        self._next_noise_value = NOISE_VALUES_SIZE
         self.reset(optimized_learning=bool(optimized_learning))
 
     def __repr__(self):
@@ -139,14 +149,16 @@ class Memory(dict):
             raise RuntimeError(f"Optimized learning cannot be enabled if the decay, {self._decay}, is not less than 1")
         if preserve_prepopulated:
             preserved = {k: v for k, v in self.items() if v._creation == 0}
-            print(preserved)
         self.clear()
         self._time = 0
         if optimized_learning is not None:
             self._optimized_learning = bool(optimized_learning)
         if preserve_prepopulated:
             for k, v in preserved.items():
-                v._references = 0 if self._optimized_learning else [0]
+                v._references = np.empty(1, dtype=int) if self._optimized_learning else np.array([0])
+                v._reference_count = 1
+                v._base_activation_time = None
+                v._base_activation = None
                 self[k] = v
         self._clear_noise_cache()
 
@@ -360,8 +372,7 @@ class Memory(dict):
             elif self._optimized_learning:
                 self._ln_1_mius_d = "illegal value" # ensure error it attempt to use this
                 raise ValueError(f"The decay, {value}, must be less than one if optimized_learning is True")
-        self._expt_cache = [None]*TRANSCENDENTAL_CACHE_SIZE
-        self._ln_cache = [None]*TRANSCENDENTAL_CACHE_SIZE
+        self._ln_cache = [None]*LN_CACHE_SIZE
         self._decay = value
         for c in self.values():
             c._clear_base_activation()
@@ -512,8 +523,53 @@ class Memory(dict):
             raise ValueError(
                 f"A value assigned to activation_history must be a MutableSequence ({value}).")
 
+    @property
+    def chunks(self):
+        """ Returns a :class:`list` of the :class:`Chunk`s contained in this :class:`Memory`.
+        """
+        return list(self.values())
+
+    def print_chunks(self, file=sys.stdout, pretty=True):
+        """Prints descriptions of all the :class:`Chunk` objects contained in this :class:`Memory`.
+        The descriptions are printed to *file*, which defaults to the standard output. If
+        *file* is not an open text file it should be a string naming a file to be opened
+        for writing.
+
+        If *pretty* is true, the default, a format intended for reading by humans is used.
+        Otherwise comma separated values (CSV) format, more suitable for importing into
+        spreadsheets, numpy, and the like, is used.
+
+        .. warning::
+            This method is intended as a debugging aid, and generally is not suitable for
+            use as a part of models.
+        """
+        if not self:
+            return
+        if isinstance(file, io.TextIOBase):
+            data = [{"chunk name": c._name,
+                     "chunk contents": dict(k).__repr__()[1:-1],
+                     "chunk created at": c._creation,
+                     "chunk references": (c.references if isinstance(c.references, Number)
+                                          else c.references.__repr__()[1:-1])}
+                    for k, c in self.items()]
+            if pretty:
+                tab = PrettyTable()
+                tab.field_names = data[0].keys()
+                for d in data:
+                    tab.add_row(d.values())
+                print(tab, file=file, flush=True)
+            else:
+                w = csv.DictWriter(file, data[0].keys())
+                w.writeheader()
+                for d in data:
+                    w.writerow(d)
+        else:
+            with open(file, "w+", newline=(None if pretty else "")) as f:
+                self.print_chunks(f, pretty)
+
     @property
     def optimized_learning(self):
+
         """A boolean indicating whether or not this Memory is configured to use optimized learning.
         Cannot be set directly, but can be changed when calling :meth:`reset`.
 
@@ -532,7 +588,7 @@ class Memory(dict):
     _similarity_cache = pylru.lrucache(SIMILARITY_CACHE_SIZE)
 
     # possibly update docstrings for retrieve() and blend() to reflect new similarity stuff,
-    # or maybe put it all n set_similarity_function() and/or mismatch?
+    # or maybe put it all in set_similarity_function() and/or mismatch?
     def _similarity(self, x, y, attribute):
         # always returns the "natural" similarity
         # returns None if similarity is inapplicatble to attribute
@@ -613,10 +669,11 @@ class Memory(dict):
         return old
 
     def _cite(self, chunk):
-        if self._optimized_learning:
-            chunk._references += 1
-        else:
-            chunk._references.append(self._time)
+        if not self._optimized_learning:
+            if chunk._reference_count >= chunk._references.size:
+                chunk._references.resize(2 * chunk._references.size, refcheck=False)
+            chunk._references[chunk._reference_count] = self._time
+        chunk._reference_count += 1
         chunk._base_activation_time = None
 
     def forget(self, when, **kwargs):
@@ -637,14 +694,19 @@ class Memory(dict):
         chunk = self.get(signature)
         if not chunk:
             return False
-        if self._optimized_learning:
-            chunk._references -= 1
-        else:
+        if not self._optimized_learning:
             try:
-                chunk._references.remove(when)
-            except ValueError:
+                i = np.where(chunk._references == when)[0][0]
+            except IndexError:
                 return False
-        if not chunk._references:
+            if i < chunk._reference_count:
+                chunk._references[i:chunk._reference_count-1] = chunk._references[i+1:chunk._reference_count]
+        elif when < chunk._creation:
+            return False
+        elif when == chunk._creation and chunk._reference_count > 1:
+            raise RuntimeError("Can't meaningfully forget a chunk at its creation time with optimized learning")
+        chunk._reference_count -= 1
+        if not chunk._reference_count:
             del self[signature]
         return True
 
@@ -716,8 +778,11 @@ class Memory(dict):
         else:
             result = None
         if result is None:
-            p = random.uniform(sys.float_info.epsilon, 1 - sys.float_info.epsilon)
-            result = self._noise * math.log((1.0 - p) / p)
+            if self._next_noise_value >= NOISE_VALUES_SIZE:
+                self._noise_values = self._rng.logistic(scale=self._noise, size=NOISE_VALUES_SIZE)
+                self._next_noise_value = 0
+            result = self._noise_values[self._next_noise_value]
+            self._next_noise_value += 1
         if self._activation_noise_cache is not None:
             self._activation_noise_cache[chunk._name] = result
         return result
@@ -964,7 +1029,7 @@ def set_similarity_function(function, *slots):
 
 class Chunk(dict):
 
-    __slots__ = ["_name", "_memory", "_creation", "_references",
+    __slots__ = ["_name", "_memory", "_creation", "_references", "_reference_count",
                  "_base_activation_time", "_base_activation"]
 
     _name_counter = 0;
@@ -975,16 +1040,30 @@ class Chunk(dict):
         self._memory = memory
         self.update(content)
         self._creation = memory._time
-        self._references = 0 if memory._optimized_learning else []
+        self._references = np.empty(1, dtype=int)
+        self._reference_count = 0
         self._base_activation_time = None
         self._base_activation = None
 
     def __repr__(self):
-        return "<Chunk {} {}>".format(self._name, dict(self))
+        return "<Chunk {} {} {}>".format(self._name, dict(self), self._reference_count)
 
     def __str__(self):
         return self._name
 
+    @property
+    def references(self):
+        """Returns when this :class:`Chunk` has been reinforced.
+        The type of the value returned depends upon whether or not the :class:`Memory`
+        containing it is using optimized learning. If it is this is an integer, the number
+        of times this :class:`Chunk` has been reinforced, and otherwise is a list of times
+        at which it was reinforced.
+        """
+        if self._memory._optimized_learning:
+            return self._reference_count
+        else:
+            return list(self._references[:self._reference_count])
+
     def _activation(self, for_partial=False):
         # Does not include the mismatch penalty component, that's handled by the caller.
         base = self._get_base_activation() if self._memory._decay is not None else 0
@@ -994,9 +1073,9 @@ class Chunk(dict):
             history = {"name": self._name,
                        "creation_time": self._creation,
                        "attributes": tuple(self.items()),
-                       "references": (self._references
+                       "references": (self._reference_count
                                       if self._memory.optimized_learning
-                                   else tuple(self._references)),
+                                      else tuple(self._references[:self._reference_count])),
                        "base_activation": base,
                        "activation_noise": noise}
             if not for_partial:
@@ -1004,47 +1083,43 @@ class Chunk(dict):
             self._memory._activation_history.append(history)
         return result
 
-    # Note that memoizing expt and ln doesn't make much difference, but it does speed
+    # Note that memoizing ln doesn't make much difference, but it does speed
     # things up a tiny bit, most noticeably under PyPy.
-    def _cached_expt(self, base):
-        try:
-            result = self._memory._expt_cache[base]
-            if result is None:
-                result = math.pow(base, -self._memory._decay)
-                self._memory._expt_cache[base] = result
-            return result
-        except (IndexError, TypeError):
-            return math.pow(base, -self._memory._decay)
-
     def _cached_ln(self, arg):
         try:
             result = self._memory._ln_cache[arg]
-            if result is None:
-                result = math.log(arg)
-                self._memory._ln_cache[arg] = result
-            return result
         except (IndexError, TypeError):
             return math.log(arg)
+        if result is None:
+            result = math.log(arg)
+            self._memory._ln_cache[arg] = result
+        return result
 
     def _get_base_activation(self):
         if self._base_activation_time != self._memory.time:
-            try:
-                if self._memory._optimized_learning:
-                    self._base_activation = (self._cached_ln(self._references)
+            err = None
+            if self._memory._optimized_learning:
+                try:
+                    self._base_activation = (self._cached_ln(self._reference_count)
                                              - self._memory._ln_1_mius_d
                                              - self._memory._decay * self._cached_ln(self._memory._time - self._creation))
-                else:
-                    base = sum(self._cached_expt(self._memory._time - ref)
-                               for ref in self._references)
+                except ValueError as e:
+                    err = e
+            else:
+                base = np.sum((self._memory._time - self._references[0:self._reference_count])
+                              ** -self._memory._decay)
+                if np.isfinite(base):
                     self._base_activation = math.log(base)
-            except ValueError as e:
+                else:
+                    err = RuntimeError(f"Non-finite value {base} encounterd when computing base activation")
+            if err:
                 if self._memory._time <= self._creation:
                     raise RuntimeError("Can't compute activation of a chunk at or before the time it was created")
                 elif (not self._memory._optimized_learning
                       and self._references[-1] >= self._memory._time):
                     raise RuntimeError("Can't compute activation of a chunk at or before the time of its most recent reference")
                 else:
-                    raise e
+                    raise err
             self._base_activation_time = self._memory.time
         return self._base_activation
 
